package jasymca;

/*
 Copyright (C) 2006, 2009 - Helmut Dersch  der@hs-furtwangen.de
                     2011       S.Chekanov

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*------------------------------------------------------------*/

import java.util.*;
import java.io.*;

///////// Start J2SE-only +

public class Jasymca implements Runnable {

	// Start interactive session
	public static void main(String args[]) {
		// get name of startupfile
		String s = System.getProperty("JASYMCA_RC");
		if (s != null)
			JasymcaRC = s;
		// Start interactive session
		String ui = "Octave";
		if (args.length > 0)
			ui = args[0];
		new Jasymca(ui).start(System.in, System.out);
	}

	public static InputStream getFileInputStream(String fname) throws IOException {
		return new FileInputStream(fname);
	}

	static OutputStream getFileOutputStream(String fname, boolean append)
			throws IOException {
		return new FileOutputStream(fname, append);
	}

	static String JasymcaRC = "Jasymca.";

	// ////// End J2SE-only -
	public void destroyApp(boolean a) {
	}

	public void pauseApp() {
	}

	// /
	// Jasymca = Java Symbolic Calculator
	// Syntax loosely related to GNU-Maxima
	// Internal Design:
	// Expressions are Lisp-lists in prefix notation
	// Elements are either parseable Lists or Algebraics
	// Variables are one of
	// --- SimpleVariable = name (String)
	// --- FunctionVariable = function (LambdaAlgebraic) + Argument (Algebraic)
	// Algebraics are one of
	// --- Zahl.Unexakt ---> double complex
	// --- Zahl.Exakt ---> BigInteger rational complex
	// --- Polynomial ---> Coefficients (Algebraic) + Variable
	// --- Polynomial.Constant ---> Coefficients (Algebraic) + immutable
	// Variable
	// --- Rational ---> Nominator (Algebraic) + Denominator (Polynomial)
	// --- Vektor ---> Components (Algebraic)
	// --- Matrix ---> Vektor of Components
	// Environment for variables, functions
	// and operators
	// Stored by name, case-(in)sensitive
	public Environment env;
	public Processor proc = null;
	public Parser pars;
	String ui = "Octave";
	// All output uses these channels
	public PrintStream ps;
	public InputStream is;

	public void interrupt() {
		if (proc != null)
			proc.set_interrupt(true);
	}

	String welcome = "%Jasymca calculator. Type \"demo\" to show examples, \"help\" for help";
	static NumFmt fmt = new NumFmtVar(10, 5);
	Thread evalLoop = null;

	public Jasymca() {
		this("Octave");
	}

	public Jasymca(String ui) {
		setup_ui(ui, true);
		welcome += "  " + ui + "-Mode.\n";
	}

	public void setup_ui(String ui, boolean clear_env) {
		if (clear_env) {
			env = new Environment();
		}
		if (ui != null)
			this.ui = ui;
		if (this.ui.equals("Maxima")) {
			// Setup environment
			proc = new XProcessor(env);
			pars = new MaximaParser(env);
		} else if (this.ui.equals("Octave")) {
			// Setup environment
			proc = new Processor(env);
			pars = new OctaveParser(env);
		} else {
			System.out.println("Mode " + this.ui + " not available.");
			System.exit(0);
		}
	}

	public void start(InputStream is, PrintStream ps) {
		this.is = is;
		this.ps = ps;
		// Read startup file

		/*
		 * try{ String fname = JasymcaRC + ui + ".rc"; InputStream file =
		 * getFileInputStream(fname); LambdaLOADFILE.readFile( file );
		 * }catch(Exception e){ ps.println("\n from Jasymca class:\n");
		 * e.printStackTrace(ps); }
		 */
		// ps.print(welcome);
		proc.setPrintStream(ps);

		// only for interactive!
		// evalLoop = new Thread(this);
		// evalLoop.start();
	}

	public void run() {
		while (true) {
			ps.print(pars.prompt()); // Prompt
			try {
				proc.set_interrupt(false);
				List code = pars.compile(is, ps);
				if (code == null) {
					ps.println("");
					continue;
				}
				if (proc.process_list(code, false) == proc.EXIT) {
					ps.println("\nGoodbye.");
					return;
				}
				proc.printStack();
			} catch (Exception e) {
				ps.println("\n from Jasymca class:\n" + e);
				proc.clearStack();
				// e.printStackTrace(ps);
			}
		}
	}
}
