package jasymca;

/*
   Copyright (C) 2006,2009 - Helmut Dersch  der@hs-furtwangen.de
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*------------------------------------------------------------*/

import java.util.*;
import java.io.*;

 import java.math.BigInteger;



// Definitions for function pointers


// Lambda is a general function pointer
public abstract class Lambda implements Constants{

 public static Processor pc;
 public static Parser pr;
 public static final boolean debug = false;
 protected static void p(String s) { if (debug) System.out.println(s); }

 protected static int length = 1; // Number of outputvalues


 protected int lambda(Stack x) throws ParseException, JasymcaException{ return 0; }


 public static Algebraic getAlgebraic( Stack st ) throws ParseException, JasymcaException{
  Object arg_in = st.pop();
  if( !(arg_in instanceof jasymca.Algebraic) ){
   pc.process_instruction( arg_in, true );
   arg_in = st.pop();
  }
  if( !(arg_in instanceof jasymca.Algebraic) )
   throw new JasymcaException("Expected algebraic, got: "+arg_in);
  return (Algebraic)arg_in;
 }

 public static Zahl getNumber(Stack st) throws ParseException, JasymcaException{
  Object arg = st.pop();
  if(arg instanceof jasymca.Algebraic)
   arg = new ExpandConstants().f_exakt((Algebraic)arg);
  if(!(arg instanceof jasymca.Zahl) )
   throw new ParseException("Expected number, got "+arg);
  return (Zahl)arg;
 }


 public static int getNarg( Stack st ) throws ParseException, JasymcaException{
  Object arg_in = st.pop();
  if( !(arg_in instanceof Integer) )
   throw new JasymcaException("Expected Integer, got: "+arg_in);
  return ((Integer)arg_in).intValue();
 }

 public static String getSymbol( Stack st ) throws ParseException, JasymcaException{
  Object arg_in = st.pop();
  if( !(arg_in instanceof java.lang.String) ||
   ((String)arg_in).length()==0 ||
   ((String)arg_in).charAt(0)==' ')
   throw new JasymcaException("Expected Symbol, got: "+arg_in);
  return (String)arg_in;
 }

 public static Polynomial getPolynomial(Stack st) throws ParseException, JasymcaException{
  Object arg = getAlgebraic( st );
  if(!(arg instanceof jasymca.Polynomial) )
   throw new ParseException("Expected polynomial, got "+arg);
  return (Polynomial)arg;
 }

 public static Vektor getVektor(Stack st) throws ParseException, JasymcaException{
  Object arg = st.pop();
  if(!(arg instanceof jasymca.Vektor) )
   throw new ParseException("Expected vector, got "+arg);
  return (Vektor)arg;
 }

 public static Variable getVariable(Stack st) throws ParseException, JasymcaException{
  Polynomial p = getPolynomial(st);
  return p.var;
 }

 public static int getInteger(Stack st) throws ParseException, JasymcaException{
  Object arg = st.pop();
  if(!(arg instanceof jasymca.Zahl) || !((Zahl)arg).integerq() )
   throw new ParseException("Expected integer, got "+arg);
  return ((Zahl)arg).intval();
 }

 public static int getInteger(Algebraic arg) throws ParseException, JasymcaException{
  if(!(arg instanceof jasymca.Zahl) || !((Zahl)arg).integerq() )
   throw new ParseException("Expected integer, got "+arg);
  return ((Zahl)arg).intval();
 }

 public static List getList(Stack st) throws ParseException, JasymcaException{
  Object arg = st.pop();
  if(!(arg instanceof java.util.List) )
   throw new ParseException("Expected list, got "+arg);
  return (List)arg;
 }


 public static Zahl ensure_Zahl( Object x )
       throws JasymcaException{
  if( !(x instanceof jasymca.Zahl) )
   throw new JasymcaException("Expected number, got "+x);
  return (Zahl)x;
 }

public  static Environment sandbox = null;
 // Insert x in rule "f(x)" and evaluate expression to Algebraic
 public static Algebraic evalx(String rule, Algebraic x) throws JasymcaException{
  try{
   List pgm = pr.compile( rule );
   Environment global = pc.getEnvironment();
   if( sandbox == null ){
    sandbox = new Environment();
    sandbox.putValue( "x", new Polynomial(new SimpleVariable("x")));
    sandbox.putValue( "X", new Polynomial(new SimpleVariable("X")));
    sandbox.putValue( "a", new Polynomial(new SimpleVariable("a")));
    sandbox.putValue( "b", new Polynomial(new SimpleVariable("b")));
    sandbox.putValue( "c", new Polynomial(new SimpleVariable("c")));
   }
   pc.setEnvironment( sandbox );
   pc.process_list( pgm, true );
   pc.setEnvironment( global );
   Algebraic y = getAlgebraic( pc.stack );
   y = y.value( new SimpleVariable("x"), x );
   return y;
  }catch(Exception e){
   throw new JasymcaException("Could not evaluate expression "+rule+": "+e.toString());
  }
 }

}

class UserProgram extends Lambda{
	// User defined functions 
 String fname;
 List body;
 SimpleVariable[] var;
 SimpleVariable result;
 Environment env=null;
 Stack ups= null;

 public UserProgram() throws ParseException,JasymcaException {}

 public UserProgram( String fname,
       SimpleVariable var[],
       List body,
       SimpleVariable result,
       Environment env,
       Stack ups){
  this.fname = fname;
  this.var = var;
  this.body = body;
  this.result= result;
  this.env = env;
  this.ups = ups;
 }

 public int lambda(Stack st) throws ParseException,JasymcaException{
  int narg = getNarg( st );
  if( var.length != narg )
   throw new JasymcaException(fname+" requires "+var.length+" Arguments.");
  // Evaluate args in global environment
  // and store values in local environment
  for(int i=0; i<var.length; i++){
   Object a = st.pop();
   env.putValue( var[i].name, a );
  }
  // process private code in sandbox
  int ret = process_block( body, ups, env, result!=null );
  if( ret != Processor.ERROR ){
   Object y = (result!=null?env.getValue( result.name ):ups.pop());
   if( y instanceof jasymca.Algebraic && result!=null)
    ((Algebraic)y).name = result.name;
   if( y!= null )
    st.push(y);
  }
  return 0;
 }

 static int process_block( List code, Stack st, Environment env, boolean clear_stack){
//System.out.println("Block: "+code+"\n"+st+"\n"+env);
  Environment global = pc.getEnvironment();
  Stack old_stack = pc.stack;
  pc.setEnvironment( env );
  pc.stack = st;
  int ret;
  try{
   ret = pc.process_list( code, true );
  }catch( Exception e ){
   ret = Processor.ERROR;
  }
  pc.stack = old_stack;
  pc.setEnvironment( global );
  if( clear_stack ){
   //st.clear();
   while( !st.empty() )
    st.pop();
  }
  return ret;
 }


}


// User defined functions 
class UserFunction extends LambdaAlgebraic{
 String fname;
 Algebraic body;
 SimpleVariable[] var;
 SimpleVariable result;
 Environment env=null;

 public UserFunction() throws ParseException,JasymcaException {}

 public UserFunction( String fname,
       SimpleVariable var[],
       Algebraic body,
       SimpleVariable result,
       Environment env){
  this.fname = fname;
  this.var = var;
  this.body = body;
  this.result= result;
  this.env = env;
 }

 public Zahl f( Zahl x ) throws JasymcaException{
  Algebraic y=f_exakt(x);
  if(y instanceof jasymca.Zahl)
    return (Zahl)y;
  y = (new ExpandConstants()).f_exakt(y);
  if(y instanceof jasymca.Zahl)
    return (Zahl)y;
  throw new JasymcaException("Can not evaluate Function "+fname+
    " to number, got "+y+" for "+x);
 }

 Algebraic f_exakt(Algebraic x) throws JasymcaException{
  if(var.length != 1)
   throw new JasymcaException("Wrong number of arguments.");
  Algebraic y = body.value(var[0], x);
  return y;
 }

 Algebraic f_exakt(Algebraic x, Algebraic y) throws JasymcaException{
  if(var.length != 2)
   throw new JasymcaException("Wrong number of arguments.");
  Algebraic z = body.value(var[0], y);
  z = z.value(var[1], x);
  return z;
 }

 Algebraic f_exakt(Algebraic[] x) throws JasymcaException{
  if(var.length != x.length)
   throw new JasymcaException("Wrong number of arguments.");
  Algebraic y = body;
  for( int i=0; i<x.length; i++)
   y = y.value(var[x.length-i-1], x[i]);
  return y;
 }



 // todo: eliminate
 Algebraic fv(Vektor x) throws JasymcaException{
  Environment global = pc.env;
  pc.env = env;
  Algebraic r = body;
  pc.env = global;
  for(int i=0; i<var.length; i++){
   r = r.value(var[i], x.get(i));
  }
  return r;
 }

 public Algebraic integrate(Algebraic arg, Variable x) throws JasymcaException{
  if( !(body instanceof jasymca.Algebraic) ){
   throw new JasymcaException("Can not integrate function "+fname);
  }
  if(!(arg.depends(x)))
   throw new JasymcaException("Expression in function does not depend on Variable.");
  if(var.length == 1)
   return ((Algebraic)body).value(var[0], arg).integrate(x);
  if(arg instanceof jasymca.Vektor && ((Vektor)arg).length()==var.length)
   return fv((Vektor)arg).integrate(x);
  throw new JasymcaException("Wrong argument to function "+fname);
 }


}




class LambdaFLOAT extends LambdaAlgebraic{
	// Convert to floating point number as much as 
	// possible, use ALGEPSILON to reduce numbers	
	
 double eps = 1.e-8;
 public int lambda(Stack st) throws ParseException, JasymcaException{
  int narg = getNarg( st );
  Algebraic exp = getAlgebraic(st);
  Zahl a = pc.env.getnum("algepsilon");
  if(a !=null){
   double epstry = a.unexakt().real;
   if(epstry>0) eps = epstry;
  }
  // Eval constants to numbers (pi etc)
  exp = new ExpandConstants().f_exakt(exp);
  st.push(exp.map(this));
  return 0;
 }
 public Zahl f(Zahl x) throws JasymcaException{
  Unexakt f = x.unexakt();
  if(f.equals(Zahl.ZERO))
   return f;
  double abs = ((Unexakt)f.abs()).real;
  double r = f.real;
  if( Math.abs(r/abs)<eps ) r = 0.;
  double i = f.imag;
  if( Math.abs(i/abs)<eps ) i = 0.;
  return new Unexakt(r,i);
 }

 Algebraic f_exakt(Algebraic x) throws JasymcaException{
  return x.map(this);
 }
}


class CreateVector extends Lambda{
	// Operator for '[' symbol
 public int lambda(Stack st) throws ParseException, JasymcaException{
  // if arg is a list of scalars or vektors of scalars:
  // create vector
  // if arg is a list of vectors of vectors, create matrix
  // if arg is a list of matrices, create matrix
  int nr = getNarg( st );
  int nrow = 1, ncol = 1;

  Matrix m = new Matrix(nr,1);
  while(nr-- > 0){
   Algebraic row = crv( st ); // insert row vector												
   Index idx = new Index( nrow, ncol, row);
   m.insert( new Matrix(row), idx);
   nrow = idx.row_max+1;
  }
  st.push( m.reduce() );
  return 0;
 }

 static Algebraic crv( Stack st ) throws ParseException, JasymcaException{
  int nc = getNarg( st );
  if(nc == 1)
   return getAlgebraic( st );
  Matrix m = new Matrix(1,nc);
  int nrow = 1, ncol = 1;
  while( nc-- > 0 ){
   Algebraic x = getAlgebraic( st );
   Index idx = new Index( nrow, ncol, x);
   m.insert( new Matrix(x), idx);
   ncol = idx.col_max+1;
  }
  return m.reduce();
 }


}


class CR1 extends Lambda{
	// Operator for 'a:b:c' symbol	
 public int lambda(Stack st) throws ParseException, JasymcaException{
  // Variable definition, store in Prefix format and return success
  int narg = getNarg( st );


  Algebraic a,b,c = getAlgebraic(st);
  if( narg==2 ){
   b = Zahl.ONE;
   a = getAlgebraic(st);
  }else{
   b = getAlgebraic(st);
   a = getAlgebraic(st);
  }
  Algebraic na = c.sub(a).div(b);
  if(!(na instanceof jasymca.Zahl)){
   na = (new ExpandConstants()).f_exakt(na);
  }
  if(!(na instanceof jasymca.Zahl))
   throw new ParseException("CreateVector requires numbers.");
  int n = (int)( ((Zahl)na).unexakt().real+1.0);
  Algebraic[] coord = new Algebraic[n];
  for(int i=0; i<n; i++)
   coord[i] = a.add(b.mult(new Unexakt((double)i)));
  st.push( new Vektor(coord) );
  return 0;
 }
}

class LambdaEXPAND extends Lambda{
 public int lambda(Stack st) throws ParseException, JasymcaException{
  int narg = getNarg( st );
  Object x = st.pop();
  if(x instanceof java.util.List){
   pc.process_list((List)x, true);
   x = pc.stack.pop();
  }
  if(x instanceof jasymca.Algebraic)
   x = new SqrtExpand().f_exakt( (Algebraic) x );
  st.push( x );
  return 0;
 }
}

class LambdaSQRT extends LambdaAlgebraic{
 public LambdaSQRT(){ diffrule = "1/(2*sqrt(x))"; intrule = "2/3*x*sqrt(x)"; }

 // Integrate root of squares
 static String intrule2 =
  "(2*a*x+b)*sqrt(X)/(4*a)+(4*a*c-b*b)/(8*a*sqrt(a))*log(2*sqrt(a*X)+2*a*x+b)";

 public Algebraic integrate(Algebraic arg, Variable x) throws JasymcaException{
  try{
   return super.integrate(arg,x);
  }catch(JasymcaException je){

  if(!(arg.depends(x)))
   throw new JasymcaException("Expression in function does not depend on Variable.");
  if( !(arg instanceof jasymca.Polynomial) || ((Polynomial)arg).degree()!=2 ||
   !((Polynomial)arg).ratfunc(x) )
   throw new JasymcaException("Can not integrate function ");

  Algebraic xp = new Polynomial(x);
  Polynomial X = (Polynomial)arg;

  Algebraic y = evalx( intrule2, xp );
  y = y.value( new SimpleVariable("X"), X );
  y = y.value( new SimpleVariable("a"), X.a[2] );
  y = y.value( new SimpleVariable("b"), X.a[1] );
  y = y.value( new SimpleVariable("c"), X.a[0] );

  y = new SqrtExpand().f_exakt(y);
  return y;
  }
 }

 public Zahl f( Zahl x) throws JasymcaException{
  Unexakt z = x.unexakt();
  if(z.imag == 0.){
   if(z.real<0.)
    return new Unexakt(0, Math.sqrt(-z.real));
   return new Unexakt(Math.sqrt(z.real));
  }
  double sr = Math.sqrt(Math.sqrt(z.real*z.real+z.imag*z.imag));
  double phi = Math.atan2(z.imag,z.real)/2.;
  return new Unexakt( sr*Math.cos(phi), sr*Math.sin(phi));
 }

 Algebraic f_exakt(Algebraic x) throws JasymcaException{
  if(x.equals(Zahl.ONE) || x.equals(Zahl.ZERO))
   return x;
  if(x.equals(Zahl.MINUS))
   return Zahl.IONE;
  if(x instanceof jasymca.Zahl){
   return fzexakt((Zahl)x);
  }

  if(x instanceof jasymca.Polynomial && ((Polynomial)x).degree()==1 &&
  ((Polynomial)x).a[0].equals(Zahl.ZERO) &&
  ((Polynomial)x).a[1].equals(Zahl.ONE) &&
  ((Polynomial)x).var instanceof jasymca.FunctionVariable &&
   ((FunctionVariable)((Polynomial)x).var).fname.equals("exp"))
   return FunctionVariable.create("exp",
    ((FunctionVariable)((Polynomial)x).var).arg.div(Zahl.TWO));
  return null;
 }


 Algebraic fzexakt(Zahl x) throws JasymcaException{
  if(x instanceof jasymca.Exakt && !x.komplexq()){
   if(x.smaller(Zahl.ZERO)){
    Algebraic r = fzexakt((Zahl)x.mult(Zahl.MINUS));
    if(r!=null) return Zahl.IONE.mult(r);
    return r;
   }
   long nom = ((Exakt)x).real[0].longValue();
   long den = ((Exakt)x).real[1].longValue();
   // x = sqrt(a/b) = sqrt((a0^2*a1)/(b0^2*b1)) = (a0/(b0*b1))*sqrt(a1*b1);
   long a0 = introot(nom), a1 = nom/(a0*a0);
   long b0 = introot(den), b1 = den/(b0*b0);
   BigInteger br[] = { BigInteger.valueOf(a0),
        BigInteger.valueOf(b0*b1) };
   Exakt r = new Exakt(br);
   a0 = a1*b1;
   if(a0 == 1L) return r;
   return r.mult( new Polynomial(new FunctionVariable("sqrt",
         new Exakt( BigInteger.valueOf(a0) ), this)));
  }
  return null;
 }

 // Find and return largest squared factor in x:
 // x = a^2*b ---> a
 long introot(long x){
  long s = 1L, f, g, t[] = { 2L, 3L, 5L};

  for(int i=0; i<t.length; i++){
   g = t[i]; f = g*g;
   while(x % f == 0L && x != 1L){
    s *= g; x /= f;
   }
  }
  for(long i= 6L; x!=1L ; i+=6L){
   g = i+1; f = g*g;
   while(x % f == 0L && x != 1L){
    s *= g; x /= f;
   }
   g = i+5; f = g*g;
   while(x % f == 0L && x != 1L){
    s *= g; x /= f;
   }
   if(f>x) break;
  }
  return s;
 }
}

// Expand all user functions
class ExpandUser extends LambdaAlgebraic{
 Algebraic f_exakt(Algebraic x1) throws JasymcaException{
  if( !(x1 instanceof jasymca.Polynomial) )
   return x1.map(this);
  Polynomial p = (Polynomial)x1;
  if(p.var instanceof jasymca.SimpleVariable)
   return p.map(this);
  FunctionVariable f = (FunctionVariable)p.var;
  Object lx = pc.env.getValue(f.fname);
  if(!(lx instanceof jasymca.UserFunction))
   return p.map(this);
  UserFunction la = (UserFunction)lx;
  if( !(la.body instanceof jasymca.Algebraic) )
   return x1;
  Algebraic body = (Algebraic)la.body;
  Algebraic x;
  if(la.var.length==1)
   x = body.value(la.var[0], f.arg);
  else if(f.arg instanceof jasymca.Vektor && ((Vektor)f.arg).length()==la.var.length)
   x = la.fv((Vektor)f.arg);
  else
   throw new JasymcaException("Wrong argument to function "+la.fname);

  Algebraic r=Zahl.ZERO;
  for(int i=p.a.length-1; i>0; i--){
   r=r.add(f_exakt(p.a[i])).mult(x);
  }
  if(p.a.length>0)
   r=r.add(f_exakt(p.a[0]));
  return r;
 }
}


// Assignment     x:3; --> assign value 3 to variable x
// assigning "null" deletes the variable
class ASS extends Lambda{
 public int lambda(Stack st) throws ParseException, JasymcaException{
  int narg = getNarg( st );

  Object val[] = new Object[narg];
  for(int i=0; i<narg; i++)
   val[i] = st.pop();
/*		
		if(narg>1){
			int nout = getNarg( st );
			if(nout != narg){
				throw new JasymcaException("Wrong number of Output variables.");
			}
		}
*/
  for(int i=narg-1; i>=0; i--){
   String name = getSymbol( st );
   if( !name.startsWith("$") )
    throw new JasymcaException("Illegal lvalue: "+name);
   name = name.substring(1);
   boolean idxq = !st.empty() &&
        st.peek() instanceof Integer;
   if( !idxq ){
    pc.env.putValue( name, val[i] );
    if(val[i] instanceof jasymca.Algebraic)
     ((Algebraic)val[i]).name=name;
   }else{
    if(!(val[i] instanceof jasymca.Algebraic)){
     throw new JasymcaException("No index allowed here: "+val[i]);
    }
    Matrix rhs = new Matrix( (Algebraic)val[i] );
    Matrix lhs = new Matrix( (Algebraic)pc.env.getValue( name ) );
    Index idx = Index.createIndex( st, lhs );
    lhs.insert( rhs, idx );
    val[i] = lhs.reduce();
    pc.env.putValue(name, val[i] );
   }
  }
  for(int i=0; i<narg; i++)
   st.push( val[i] );
  return 0;
 }

 static int lambdap(Stack st, Lambda op) throws ParseException, JasymcaException{
  int narg = getNarg( st );
  Object y = st.pop();
  String name = getSymbol( st );
  if( !name.startsWith("$") )
   throw new JasymcaException("Illegal lvalue: "+name);
  List t = Comp.vec2list( new Vector() );
  t.add( name );
  t.add( name.substring(1) );
  t.add( y );
  t.add( new Integer(2) );
  t.add( op );
  t.add( new Integer(1) );
  t.add( Operator.get("=").getLambda() );
  pc.process_list( t, true );
  return 0;
 }

 static int lambdai(Stack st, boolean sign, boolean pre) throws ParseException, JasymcaException{
  int narg = getNarg( st );
  String name = getSymbol( st );
  if( !name.startsWith("$") )
   throw new JasymcaException("Illegal lvalue: "+name);
  Object p = null;
  if(!pre)
   p = pc.env.getValue( name.substring(1) );
  List t = Comp.vec2list( new Vector());
  t.add( name );
  t.add( name.substring(1) );
  t.add( Zahl.ONE );
  t.add( new Integer(2) );
  t.add( sign ? Operator.get("+").getLambda() :
        Operator.get("-").getLambda());
  t.add( new Integer(1) );
  t.add( Operator.get("=").getLambda() );
  pc.process_list( t, true );
  if(!pre && p != null){
   if( p instanceof jasymca.Algebraic )
    ((Algebraic)p).name = null;
   st.pop();
   st.push( p );
  }
  return 0;
 }

}

