package jasymca;

class LambdaEXP extends LambdaAlgebraic{
 public LambdaEXP(){ diffrule = "exp(x)"; intrule = "exp(x)"; }

 Zahl f( Zahl x){
  Unexakt z = x.unexakt();
  double r = Math.exp(z.real);
  if(z.imag!=0.)
   return new Unexakt( r*Math.cos(z.imag) , r*Math.sin(z.imag) );
  return new Unexakt(r);
 }

 Algebraic f_exakt(Algebraic x) throws JasymcaException{
  if(x.equals(Zahl.ZERO))
   return Zahl.ONE;
  if(x instanceof jasymca.Polynomial && ((Polynomial)x).degree()==1 &&
  ((Polynomial)x).a[0].equals(Zahl.ZERO) ){
   Polynomial xp = (Polynomial)x;
   if(xp.var instanceof jasymca.SimpleVariable &&
    ((SimpleVariable)xp.var).name.equals("pi")){ // exp(n*i*pi)
    Algebraic q = xp.a[1].div(Zahl.IONE);
    if(q instanceof jasymca.Zahl){
     return fzexakt((Zahl)q);
    }
   }
   if(xp.a[1] instanceof jasymca.Zahl &&
    xp.var instanceof jasymca.FunctionVariable &&
    ((FunctionVariable)xp.var).fname.equals("log")){
     if( ((Zahl)xp.a[1]).integerq()) { // exp(n*log(x)) --> x^n
      int n = ((Zahl)xp.a[1]).intval();
      return ((FunctionVariable)xp.var).arg.pow_n(n);
     }
   }
  }
  return null;
 }

 // Calculate exp(x*i*pi)
 Algebraic fzexakt(Zahl x) throws JasymcaException{
  if(x.smaller(Zahl.ZERO)){
   Algebraic r = fzexakt((Zahl)x.mult(Zahl.MINUS));
   if(r!=null) return r.cc();
   return r;
  }
  if( x.integerq() ){
   if( x.intval() % 2 == 0 ){ // n even
    return Zahl.ONE;
   }else // odd
    return Zahl.MINUS;
  }
  Algebraic qs = x.add(new Unexakt(.5));
  if( ((Zahl)qs).integerq() ){
   if( ((Zahl)qs).intval() % 2 == 0 ){ // n even
    return Zahl.IMINUS;
   }else{ // odd
    return Zahl.IONE;
    }
  }
  qs = x.mult(new Unexakt(4));
  if( ((Zahl)qs).integerq() ){
   Algebraic sq2 = FunctionVariable.create("sqrt",new Unexakt(0.5));
   switch( ((Zahl)qs).intval() % 8 ){
    case 1: return Zahl.ONE.add(Zahl.IONE).div(Zahl.SQRT2);
    case 3: return Zahl.MINUS.add(Zahl.IONE).div(Zahl.SQRT2);
    case 5: return Zahl.MINUS.add(Zahl.IMINUS).div(Zahl.SQRT2);
    case 7: return Zahl.ONE.add(Zahl.IMINUS).div(Zahl.SQRT2);
   }
  }
  qs = x.mult(new Unexakt(6));
  if( ((Zahl)qs).integerq() ){
   switch( ((Zahl)qs).intval() % 12 ){
    case 1: return Zahl.SQRT3.add(Zahl.IONE).div(Zahl.TWO);
    case 2: return Zahl.ONE.add(Zahl.SQRT3.mult(Zahl.IONE)).div(Zahl.TWO);
    case 4: return Zahl.SQRT3.mult(Zahl.IONE).add(Zahl.MINUS).div(Zahl.TWO);
    case 5: return Zahl.IONE.sub(Zahl.SQRT3).div(Zahl.TWO);
    case 7: return Zahl.IMINUS.sub(Zahl.SQRT3).div(Zahl.TWO);
    case 8: return Zahl.SQRT3.mult(Zahl.IMINUS).sub(Zahl.ONE).div(Zahl.TWO);
    case 10:return Zahl.SQRT3.mult(Zahl.IMINUS).add(Zahl.ONE).div(Zahl.TWO);
    case 11:return Zahl.IMINUS.add(Zahl.SQRT3).div(Zahl.TWO);
   }
  }
  return null;
 }
}


/*
@GROUP
core
@SYNTAX
exp(value)
@DOC
Calculates the exponent of a complex number
@EXAMPLES
exp(-5.5) = -5
exp(2.3)   = 3
@SEE
log, ln
*/
