package jasymca;

/*
 Copyright (C) 2006 - Helmut Dersch  der@hs-furtwangen.de

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  

 */

import java.util.*;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.text.*;
import net.sf.epsgraphics.*;

public class PlotGraph extends Frame implements Runnable, ComponentListener,
		MouseListener, MouseMotionListener {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	Vector PlotLines = new Vector();
	public final static int LINEAR = 0;
	final static int LOGLIN = 1;
	final static int LINLOG = 2;
	final static int LOGLOG = 3;
	int plotmode = LINEAR;

	NumFmt fmt = new NumFmtVar(10, 4);
	// Pointer parameters
	int xp = -1, yp;
	int dx = 0, dy = 0;
	double a0, a1;
	boolean movePointer = false;
	Thread moveP = null;
	static final int PMODE_POINT = 0;
	static final int PMODE_LINE = 1;
	static final int PMODE_LINE_POINT = 2;
	int pointerMode = PMODE_POINT;
	double minx = Double.POSITIVE_INFINITY;
	double maxx = Double.NEGATIVE_INFINITY;
	int ntx = 10, nty = 10;
	double miny = Double.POSITIVE_INFINITY;
	double maxy = Double.NEGATIVE_INFINITY;
	Image ShowLine = null;

	// Layoutparameter
	// window area
	int width, height;
	// width of frame around plot area
	int xleft, ytop, xright, ybottom;
	int yXlabel; // y-position of X-label
	int yTlabel; // y-position of title
	String Xlabel = null;
	String Ylabel = null;
	String Tlabel = null;

	// Colors
	static final Color BLACK = Color.black;
	static final Color RED = Color.red;
	static final Color GREEN = new Color(0x228b22); // forestgreen
	static final Color BLUE = Color.blue;
	static final Color CYAN = Color.cyan;
	static final Color MAGENTA = Color.magenta;
	static final Color YELLOW = Color.yellow;
	static final Color WHITE = Color.white;
	static final Color LIGHTGRAY = new Color(0xc0c0c0);

	boolean Hold_b = false;

	static final Stroke DOTTED = new BasicStroke(1f, BasicStroke.CAP_ROUND,
			BasicStroke.JOIN_ROUND, 1f, new float[] { 2f }, 0f);
	static final Stroke DASHED = new BasicStroke(1.0f, BasicStroke.CAP_BUTT,
			BasicStroke.JOIN_MITER, 10f, new float[] { 10f }, 0f);
	static final Stroke SOLID = new BasicStroke();
	static final Font fontAxis = new Font("Arial", Font.BOLD, 14);

	public void setmode(int mode) {
		if (mode != plotmode)
			reset();
		plotmode = mode;
	}

	void setparam() {
		width = getWidth();
		height = getHeight();
		xleft = Ylabel == null ? 50 : 70;
		xright = 25;
		ytop = Tlabel == null ? 40 : 60;
		ybottom = Xlabel == null ? 30 : 50;
	}

	public void setXlabel(String s) {
		Xlabel = s;
		repaint();
	}

	public void setYlabel(String s) {
		Ylabel = s;
		repaint();
	}

	public void setTlabel(String s) {
		Tlabel = s;
		repaint();
	}

	public void run() {
		while (movePointer) {
			try {
				if (xp < 0) {
					movePointer = false;
					return;
				}
				int xneu = xp + dx;
				if (xneu >= getScreenX(maxx))
					xneu = getScreenX(minx);
				if (xneu < getScreenX(minx))
					xneu = getScreenX(maxx);
				int yneu = yp + dy;
				if (yneu >= getScreenY(miny))
					yneu = getScreenY(maxy);
				if (yneu < getScreenY(maxy))
					yneu = getScreenY(miny);
				xp = xneu;
				yp = yneu;
				repaint();
				Thread.sleep(100);
			} catch (Exception e) {
			}
		}
	}

	public PlotGraph(int mode) {
		plotmode = mode;
		setSize(600, 400);
		setparam();
		try {
			final Frame f = this;
			addWindowListener(new WindowAdapter() {
				public void windowClosing(WindowEvent e) {
					f.setVisible(false);
				}
			});
			addComponentListener(this);
			addMouseListener(this);
			addMouseMotionListener(this);
			java.net.URL url = getClass().getResource("icons/fav.png");
			ShowLine = Toolkit.getDefaultToolkit().createImage(url);
		} catch (Exception e) {
			System.out.println("Could not load Images.");
		}
	}

	public PlotGraph() {
		this(LINEAR);
	}

	/**
	 * Draw a line
	 * 
	 * @param g
	 */
	void drawStraightLine(Graphics g) {
		// Left Point
		double xl = minx;
		double yl = a1 * xl + a0;
		if (yl > maxy) {
			yl = maxy;
			xl = (yl - a0) / a1;
		} else if (yl < miny) {
			yl = miny;
			xl = (yl - a0) / a1;
		}
		// Right Point
		double xr = maxx;
		double yr = a1 * xr + a0;
		if (yr > maxy) {
			yr = maxy;
			xr = (yr - a0) / a1;
		} else if (yr < miny) {
			yr = miny;
			xr = (yr - a0) / a1;
		}
		if (xp >= 0 && pointerMode == PMODE_LINE) { // Recalculate a0,a1
			double xm = getXCoordinate(xp);
			double ym = getYCoordinate(yp);
			int Xr = getScreenX(xr), Yr = getScreenY(yr), Xl = getScreenX(xl), Yl = getScreenY(yl);
			if ((xp - Xr) * (xp - Xr) + (yp - Yr) * (yp - Yr) < (xp - Xl)
					* (xp - Xl) + (yp - Yl) * (yp - Yl)) {
				a1 = (ym - yl) / (xm - xl);
				a0 = yl - a1 * xl;
				xr = maxx;
				yr = a1 * xr + a0;
				if (yr > maxy) {
					yr = maxy;
					xr = (yr - a0) / a1;
				} else if (yr < miny) {
					yr = miny;
					xr = (yr - a0) / a1;
				}
			} else {
				a1 = (yr - ym) / (xr - xm);
				a0 = yr - a1 * xr;
				xl = minx;
				yl = a1 * xl + a0;
				if (yl > maxy) {
					yl = maxy;
					xl = (yl - a0) / a1;
				} else if (yl < miny) {
					yl = miny;
					xl = (yl - a0) / a1;
				}
			}
		}
		// Finally, draw the line
		g.setColor(GREEN);
		g.drawLine(getScreenX(xl), getScreenY(yl), getScreenX(xr),
				getScreenY(yr));
		g.setColor(BLACK);
		// Display Coordinate Message
		if (pointerMode == PMODE_LINE) {
			drawMessage(g, xp, yp, new String[] { "a1=" + fmt.toString(a1),
					"a0=" + fmt.toString(a0) });
		}
	}

	/**
	 * Draw a message
	 * 
	 * @param g
	 * @param xp
	 * @param yp
	 * @param msg
	 */
	void drawMessage(Graphics g, int xp, int yp, String[] msg) {
		// get width of widest string
		double[] sw = new double[msg.length];
		for (int i = 0; i < sw.length; i++)
			sw[i] = ((int) Math
					.round(g.getFont()
							.getStringBounds(msg[i],
									((Graphics2D) g).getFontRenderContext())
							.getWidth()));
		int fh = ((int) Math.round(g.getFont()
				.getStringBounds("A", ((Graphics2D) g).getFontRenderContext())
				.getHeight()));
		int mw = (int) max(sw) + 10, // size of message board
		mh = msg.length * fh + 6;
		// Position (opposite to xp,yp
		int xw = xp < width / 2 ? width - xright - mw : xleft;
		int yw = yp < height / 2 ? height - ybottom - mh : ytop;
		g.setColor(LIGHTGRAY);
		g.fillRect(xw, yw, mw, mh);
		g.setColor(BLACK);
		for (int i = 0; i < msg.length; i++) {
			g.drawString(msg[i], xw + 5, yw + (i + 1) * fh + 3);
		}
	}

	/**
	 * Draw a point
	 * 
	 * @param g
	 */
	void drawPointer(Graphics g) {
		// Pointer active, draw red circle
		g.setColor(RED);
		fillCircle(g, xp, yp, 3);
		g.setColor(LIGHTGRAY);
		g.fillRect(xp, yp, 1, 1);
		double X = getXCoordinate(xp);
		if (plotmode == LOGLIN || plotmode == LOGLOG)
			X = Math.pow(10.0, X);
		double Y = getYCoordinate(yp);
		if (plotmode == LINLOG || plotmode == LOGLOG)
			Y = Math.pow(10.0, Y);
		drawMessage(g, xp, yp, new String[] { "X=" + fmt.toString(X),
				"Y=" + fmt.toString(Y) });
	}

	// Math Coordinates
	public double getXCoordinate(int x) {
		double a = (maxx - minx) / (double) (width - xleft - xright);
		double b = minx - a * xleft;
		return a * x + b;
	}

	public double getYCoordinate(int y) {
		double a = (miny - maxy) / (double) (height - ytop - ybottom);
		double b = maxy - a * ytop;
		return a * y + b;
	}

	// Screen Coordinates
	public int getScreenX(double x) {
		double a = (width - xleft - xright) / (maxx - minx);
		double b = xleft - a * minx;
		return (int) (a * x + b + 0.5);
	}

	public int getScreenY(double y) {
		double a = (height - ytop - ybottom) / (miny - maxy);
		double b = ytop - a * maxy;
		return (int) (a * y + b + 0.5);
	}

	class PlotLine {
		double lineMinx, lineMaxx, lineMiny, lineMaxy;
		char marker = ' ';
		double x[];
		double y[];
		double eu[] = null; // error up
		double el[] = null; // error down
		Color color = BLUE;

		void setPoints(double[] xp, double[] yp) {
			x = xp;
			y = yp;
			lineMaxx = max(x);
			lineMinx = min(x);
			lineMaxy = max(y);
			lineMiny = min(y);
		}

		void paint(Graphics g) {
			Color old = g.getColor();
			g.setColor(color);
			if (marker == ' ') {
				for (int i = 0; i < x.length - 1; i++)
					g.drawLine(getScreenX(x[i]), getScreenY(y[i]),
							getScreenX(x[i + 1]), getScreenY(y[i + 1]));
			} else { // draw symbols
				for (int i = 0; i < x.length - 1; i++)
					drawSymbol(g, getScreenX(x[i]), getScreenY(y[i]), marker);
			}
			if (eu != null) { // errorbars
				for (int i = 0; i < x.length - 1; i++) {
					g.drawLine(getScreenX(x[i]), getScreenY(y[i] - el[i]),
							getScreenX(x[i]), getScreenY(y[i] + eu[i]));
					g.drawLine(getScreenX(x[i]) - 3, getScreenY(y[i] - el[i]),
							getScreenX(x[i]) + 3, getScreenY(y[i] - el[i]));
					g.drawLine(getScreenX(x[i]) - 3, getScreenY(y[i] + eu[i]),
							getScreenX(x[i]) + 3, getScreenY(y[i] + eu[i]));
				}
			}
			g.setColor(old);
		}

		void drawSymbol(Graphics g, int x, int y, char marker) {
			switch (marker) {
			case '+':
				g.drawLine(x, y - 3, x, y + 3);
				g.drawLine(x - 3, y, x + 3, y);
				break;
			case 'o':
				g.drawLine(x - 1, y - 3, x + 1, y - 3);
				g.drawLine(x + 1, y - 3, x + 3, y - 1);
				g.drawLine(x + 3, y - 1, x + 3, y + 1);
				g.drawLine(x + 3, y + 1, x + 1, y + 3);
				g.drawLine(x + 1, y + 3, x - 1, y + 3);
				g.drawLine(x - 1, y + 3, x - 3, y + 1);
				g.drawLine(x - 3, y + 1, x - 3, y - 1);
				g.drawLine(x - 3, y - 1, x - 1, y - 3);
				break;
			case 'x':
				g.drawLine(x - 3, y - 3, x + 3, y + 3);
				g.drawLine(x + 3, y - 3, x - 3, y + 3);
				break;
			case '*':
			default:
				g.drawLine(x, y - 3, x, y + 3);
				g.drawLine(x - 3, y, x + 3, y);
				g.drawLine(x - 2, y - 2, x + 2, y + 2);
				g.drawLine(x + 2, y - 2, x - 2, y + 2);
				break;
			}
		}

		public void setLineAttributes(String options) {
			for (int i = 0; i < options.length(); i++) {
				switch (options.charAt(i)) {
				case 'r':
					color = RED;
					break;
				case 'g':
					color = GREEN;
					break;
				case 'b':
					color = BLUE;
					break;
				case 'y':
					color = YELLOW;
					break;
				case 'm':
					color = MAGENTA;
					break;
				case 'c':
					color = CYAN;
					break;
				case 'w':
					color = WHITE;
					break;
				case 'k':
					color = BLACK;
					break;
				default:
					marker = options.charAt(i);
				}
			}
		}
	}

	private void reset() {
		PlotLines.removeAllElements();
		minx = Double.POSITIVE_INFINITY;
		maxx = Double.NEGATIVE_INFINITY;
		miny = Double.POSITIVE_INFINITY;
		maxy = Double.NEGATIVE_INFINITY;
		Xlabel = null;
		Ylabel = null;
		Tlabel = null;
	}

	// in-place conversion
	void log10(double[] x) throws JasymcaException {
		for (int i = 0; i < x.length; i++) {
			if (x[i] <= 0.0)
				throw new JasymcaException("Log from negative number.");
			x[i] = Math.log(x[i]) / Math.log(10.0);
		}
	}

	public void addLine(Object[] params) throws JasymcaException {
		if (!Hold_b) {
			reset();
		}
		for (int i = 0; i < params.length;) {
			PlotLine line = new PlotLine(); // New Line
			if (i < params.length - 1
					&& !(params[i + 1] instanceof java.lang.String)) {
				double x[] = (double[]) params[i];
				if (plotmode == LOGLIN || plotmode == LOGLOG)
					log10(x);
				line.lineMaxx = Math.max(max(x), maxx);
				line.lineMinx = Math.min(min(x), minx);
				double y[] = (double[]) params[i + 1];
				if (plotmode == LINLOG || plotmode == LOGLOG)
					log10(y);
				line.lineMaxy = Math.max(max(y), maxy);
				line.lineMiny = Math.min(min(y), miny);
				if (x.length != y.length)
					throw new JasymcaException("X and Y must be same length");
				line.setPoints(x, y);
				i += 2;
			} else {
				double x[] = (double[]) params[i];
				if (plotmode == LINLOG || plotmode == LOGLOG)
					log10(x);
				maxx = x.length;
				minx = 1;
				maxy = max(x);
				miny = min(x);
				line.x = new double[x.length];
				for (int ind = 0; ind < x.length; ind++)
					line.x[ind] = ind + 1;
				line.setPoints(line.x, x);
				i++;
			}
			maxx = Math.max(line.lineMaxx, maxx);
			minx = Math.min(line.lineMinx, minx);
			maxy = Math.max(line.lineMaxy, maxy);
			miny = Math.min(line.lineMiny, miny);
			if (i < params.length && (params[i] instanceof java.lang.String)) {
				line.setLineAttributes(params[i].toString());
				i++;
			}
			PlotLines.addElement(line);
		}
		setMinMax();
		repaint();
	}

	void addLineErrorbars(Object[] params) throws JasymcaException {
		if (!Hold_b) {
			reset();
		}
		if (params.length < 3)
			throw new JasymcaException("At least 3 arguments required.");
		PlotLine line = new PlotLine(); // New Line
		double x[] = (double[]) params[0];
		line.lineMaxx = Math.max(max(x), maxx);
		line.lineMinx = Math.min(min(x), minx);
		double y[] = (double[]) params[1];
		line.lineMaxy = Math.max(max(y), maxy);
		line.lineMiny = Math.min(min(y), miny);
		if (x.length != y.length)
			throw new JasymcaException("X and Y must be same length");
		line.setPoints(x, y);
		double el[] = (double[]) params[2], eu[] = el;
		if (el.length != y.length)
			throw new JasymcaException("Errors and Y must be same length");
		int i = 3;
		if (params.length > 3 && !(params[3] instanceof java.lang.String)) {
			eu = (double[]) params[3];
			if (eu.length != y.length)
				throw new JasymcaException("Errors and Y must be same length");
			i++;
		}
		line.eu = eu;
		line.el = el;
		maxx = Math.max(line.lineMaxx, maxx);
		minx = Math.min(line.lineMinx, minx);
		maxy = Math.max(line.lineMaxy, maxy);
		miny = Math.min(line.lineMiny, miny);
		if (i < params.length && (params[i] instanceof java.lang.String)) {
			line.setLineAttributes(params[i].toString());
			i++;
		}
		PlotLines.addElement(line);
		setMinMax();
		repaint();
	}

	void drawGraph(Graphics g) {
		Color old = g.getColor();
		g.setColor(WHITE);
		g.fillRect(0, 0, width, height);
		for (int i = 0; i < PlotLines.size(); i++) {
			PlotLine line = (PlotLine) PlotLines.elementAt(i);
			line.paint(g);
		}
		g.setColor(old);
		// axis
		drawAxis(g);
	}

	public void update(Graphics g) {
		paint(g);
	}

	public void paint(Graphics g) {
		setparam();
		if (width < 100 || height < 100)
			return;
		drawGraph(g);
		if (pointerMode != PMODE_POINT)
			drawStraightLine(g);
		if (pointerMode != PMODE_LINE && xp >= 0)
			drawPointer(g);
		g.drawImage(ShowLine, width - xright + 3, ytop, this);
	}

	boolean inShowLine(int x, int y) {
		if (ShowLine != null) {
			if (x > width - xright + 3
					&& x < width - xright + 3 + ShowLine.getWidth(this)
					&& y > ytop && y < ytop + ShowLine.getHeight(this))
				return true;
		}
		return false;
	}

	public void fillCircle(Graphics g, int x, int y, int r) {
		int r2 = r * r;
		for (int yd = y - r; yd < y + r; yd++) {
			if (yd >= 0 && yd < height) {
				int y2 = (yd - y) * (yd - y);
				for (int xd = x - r; xd < x + r; xd++) {
					if (xd >= 0 && xd < width) {
						if (y2 + (xd - x) * (xd - x) <= r2)
							g.fillRect(xd, yd, 1, 1);
					}
				}
			}
		}
	}

	double max(double[] x) {
		double max = x[0];
		for (int i = 1; i < x.length; i++) {
			if (x[i] > max)
				max = x[i];
		}
		return max;
	}

	double min(double[] x) {
		double min = x[0];
		for (int i = 1; i < x.length; i++) {
			if (x[i] < min)
				min = x[i];
		}
		return min;
	}

	/**
	 * Draw axis
	 * 
	 * @param g
	 */
	void drawAxis(Graphics g) {
		g.setColor(BLACK);
		// Draw plotframe
		int x2 = width - xright, y2 = height - ybottom;
		g.drawLine(xleft, ytop, x2, ytop);
		g.drawLine(xleft, ytop, xleft, y2);
		g.drawLine(x2, ytop, x2, y2);
		g.drawLine(xleft, y2, x2, y2);
		if (minx < 0 && 0 < maxx) { // Draw y-Axis
			((Graphics2D) g).setStroke(DOTTED);
			;
			// ((Graphics2D)g).setStroke( DOTTED );
			g.drawLine(getScreenX(0.), getScreenY(miny), getScreenX(0.),
					getScreenY(maxy));
			((Graphics2D) g).setStroke(SOLID);
			;
		}
		if (miny < 0 && 0 < maxy) { // Draw x-Axis
			((Graphics2D) g).setStroke(DOTTED);
			;
			g.drawLine(getScreenX(minx), getScreenY(0.), getScreenX(maxx),
					getScreenY(0.));
			((Graphics2D) g).setStroke(SOLID);
			;
		}
		drawOrnaments(g);
	}

	// decimal exponent
	static int decExp(double x) {
		return (int) (Math.log(x) / Math.log(10.));
	}

	// return largest p=10^n with p<x
	double largestp10(double x) {
		double p = 1.0;
		while (p < x)
			p *= 10.0;
		while (p > x)
			p /= 10.0;
		return p;
	}

	// Set min and max so that axis-length = n * 10^a
	// mit 1 <= n <= 10
	// also, set a1 and a0
	void setMinMax() {
		if (maxx == minx)
			maxx++;
		double div = largestp10(maxx - minx) / 10;
		int ntx1 = (int) Math.ceil(maxx / div);
		int ntx2 = (int) Math.floor(minx / div);
		ntx = ntx1 - ntx2;
		maxx = ntx1 * div;
		minx = ntx2 * div;
		if (maxy == miny)
			maxy++;
		div = largestp10(maxy - miny) / 10;
		int nty1 = (int) Math.ceil(maxy / div);
		int nty2 = (int) Math.floor(miny / div);
		nty = nty1 - nty2;
		maxy = nty1 * div;
		miny = nty2 * div;
		a1 = 0.;
		a0 = (maxy + miny) / 2.;
	}

	/**
	 * Draw box
	 * 
	 * @param g
	 */
	void drawOrnaments(Graphics g) {

		g.setFont(fontAxis);

		int lenMajor = height / 40, i; // Length of Ticmarks
		double x = 0, y = 0;
		int axis_w = width - xleft - xright;
		int axis_h = height - ytop - ybottom;
		String label;
		int lenlabel = ((int) Math.round(g
				.getFont()
				.getStringBounds(" 00.00 ",
						((Graphics2D) g).getFontRenderContext()).getWidth()));
		int maxnumtics = axis_w / lenlabel;
		double dX = (maxx - minx) / ntx;
		double startx = minx;
		while ((maxx - minx) / dX > maxnumtics) {
			dX *= 2;
			startx = (int) (startx / dX + 0.5) * dX;
			if ((maxx - minx) / dX > maxnumtics) {
				dX *= 2.5;
				startx = (int) (startx / dX + 0.5) * dX;
			}
		}
		if (plotmode == LOGLIN || plotmode == LOGLOG) {
			while (Math.abs((startx / dX) - Math.round(startx / dX)) > 0.01)
				startx += dX;
			if (dX < 1.0)
				dX = 1.0;
			else if (dX == 2.5)
				dX = 2.0;
			startx = dX * Math.floor(startx / dX);
		}
		int exponent = Math.max(decExp(maxx), decExp(minx)) - 1;
		if (Math.abs(exponent) < 2)
			exponent = 0;
		double scf = Math.pow(10., (double) exponent);

		for (i = 0, x = startx; x <= maxx; i++, x += dX) {

			// fix -5.551115123125783E-17 values instead of 0
			if (Math.abs(x) < 10E-16 && dX > 10E-15)
				x = 0.0;

			int xworld = getScreenX(x);
			if (xworld < xleft)
				continue;
			// draw ticmarks
			g.drawLine(xworld, ytop, xworld, ytop + lenMajor);
			g.drawLine(xworld, height - ybottom - lenMajor, xworld, height
					- ybottom);
			if ((plotmode == LOGLIN || plotmode == LOGLOG) && (dX < 1.5)) {
				for (int k = 2; k <= 9; k++) {
					int xk = getScreenX(x + Math.log((double) k)
							/ Math.log(10.));
					if (xk > xleft && xk < xleft + width) {
						g.drawLine(xk, ytop, xk, ytop + lenMajor / 2);
						g.drawLine(xk, height - ybottom - lenMajor / 2, xk,
								height - ybottom);
					}
				}
			}
			if (plotmode == LOGLIN || plotmode == LOGLOG)
				label = "10^";
			else
				label = "";
			// draw label
			if (x + dX > maxx && exponent != 0) {
				label = "E" + exponent;
			} else {
				label += fmt.toString(x / scf);
			}
			centerText(
					g,
					label,
					height
							- ybottom
							+ 1.5
							* ((int) Math.round(g
									.getFont()
									.getStringBounds(
											"A",
											((Graphics2D) g)
													.getFontRenderContext())
									.getHeight())), xworld);
		}
		if (Xlabel != null)
			centerText(
					g,
					Xlabel,
					height
							- ybottom
							+ 3
							* ((int) Math.round(g
									.getFont()
									.getStringBounds(
											"A",
											((Graphics2D) g)
													.getFontRenderContext())
									.getHeight())), xleft
							+ (width - xleft - xright) / 2);
		if (Tlabel != null)
			centerText(
					g,
					Tlabel,
					ytop
							- ((int) Math.round(g
									.getFont()
									.getStringBounds(
											"A",
											((Graphics2D) g)
													.getFontRenderContext())
									.getHeight())) / 2, xleft
							+ (width - xleft - xright) / 2);
		if (Ylabel != null)
			centerTextV(
					g,
					Ylabel,
					((int) Math.round(g
							.getFont()
							.getStringBounds("A",
									((Graphics2D) g).getFontRenderContext())
							.getHeight())), ytop + (height - ytop - ybottom)
							/ 2);
		// Vertical unit dY
		int hilabel = ((int) Math.round(g.getFont()
				.getStringBounds("A", ((Graphics2D) g).getFontRenderContext())
				.getHeight()));
		maxnumtics = (int) ((axis_h) / (1.5 * hilabel));
		double dY = (maxy - miny) / nty;
		double starty = miny;
		while ((maxy - miny) / dY > maxnumtics) {
			dY *= 2;
			starty = (int) (starty / dY + 0.5) * dY;
			if ((maxy - miny) / dY > maxnumtics) {
				dY *= 2.5;
				starty = (int) (starty / dY + 0.5) * dY;
			}
		}
		if (plotmode == LINLOG || plotmode == LOGLOG) {
			if (dY < 1.0)
				dY = 1.0;
			else if (dY == 2.5)
				dY = 2.0;
			starty = dY * Math.floor(starty / dY);
		}
		exponent = Math.max(decExp(maxy), decExp(miny)) - 1;
		if (Math.abs(exponent) < 2)
			exponent = 0;
		scf = Math.pow(10., (double) exponent);

		for (i = 0, y = starty; y <= maxy; i++, y += dY) {

			// fix -5.551115123125783E-17 values instead of 0
			if (Math.abs(y) < 10E-16 && dY > 10E-15)
				y = 0.0;

			int yworld = getScreenY(y);
			if (yworld > height - ybottom)
				continue;
			g.drawLine(xleft, yworld, xleft + lenMajor, yworld);
			g.drawLine(xleft + axis_w, yworld, xleft + axis_w - lenMajor,
					yworld);
			if ((plotmode == LINLOG || plotmode == LOGLOG) && dY < 1.5) {
				for (int k = 2; k <= 9; k++) {
					int yk = getScreenY(y + Math.log((double) k)
							/ Math.log(10.));
					if (yk > ytop && yk < ytop + height) {
						g.drawLine(xleft, yk, xleft + lenMajor / 2, yk);
						g.drawLine(xleft + axis_w, yk, xleft + axis_w
								- lenMajor / 2, yk);
					}
				}
			}

			if (plotmode == LINLOG || plotmode == LOGLOG)
				label = "10^";
			else
				label = "";
			// draw label
			if (y + dY > maxy && exponent != 0) {
				label = "E" + exponent;
			} else {
				label += fmt.toString(y / scf);
			}

			// System.out.println(y); // chekanov
			g.drawString(
					label,
					xleft
							- ((int) Math.round(g
									.getFont()
									.getStringBounds(
											label,
											((Graphics2D) g)
													.getFontRenderContext())
									.getWidth())) - 3, yworld + hilabel / 2 - 3);
		}
	}

	public static void centerText(Graphics g, String msg, double posY,
			double centerPos) {
		int sw = ((int) Math.round(g.getFont()
				.getStringBounds(msg, ((Graphics2D) g).getFontRenderContext())
				.getWidth()));
		double posX = (centerPos) - sw / 2;
		double dy = 4., dx = 0.;
		g.drawString(msg, (int) Math.round(posX - dx),
				(int) Math.round(posY - dy));
		// return bounds;
	}

	public static void centerTextV(Graphics g, String msg, double posX,
			double centerPos) {
		int sw = ((int) Math.round(g.getFont()
				.getStringBounds(msg, ((Graphics2D) g).getFontRenderContext())
				.getWidth()));
		double posY = (centerPos) + sw / 2;
		double dy = 0., dx = 0.;
		Font oldFont = g.getFont();
		Font f = oldFont.deriveFont(java.awt.geom.AffineTransform
				.getRotateInstance(-Math.PI / 2.0));
		g.setFont(f);
		g.drawString(msg, (int) posX, (int) posY);
		g.setFont(oldFont);
	}

	void startMovePointer() {
		if (moveP != null && moveP.isAlive())
			return;
		movePointer = true;
		moveP = new Thread(this);
		moveP.start();
	}

	public void componentResized(ComponentEvent e) {
		repaint();
	}

	public void componentMoved(ComponentEvent e) {
	}

	public void componentShown(ComponentEvent e) {
	}

	public void componentHidden(ComponentEvent e) {
	}

	public void mouseClicked(MouseEvent e) {
	}

	public void mousePressed(MouseEvent e) {
		if (inShowLine(e.getX(), e.getY())) {
			xp = -1;
			if (pointerMode == PMODE_POINT)
				pointerMode = PMODE_LINE;
			else if (pointerMode == PMODE_LINE)
				pointerMode = PMODE_LINE_POINT;
			else
				pointerMode = PMODE_POINT;
		} else {
			xp = e.getX();
			yp = e.getY();
		}
		repaint();
	}

	public void mouseReleased(MouseEvent e) {
		xp = -1;
		repaint();
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	public void mouseDragged(MouseEvent e) {
		if (xp != -1) { // Entprellen
			xp = e.getX();
			yp = e.getY();
			repaint();
		}
	}

	public void mouseMoved(MouseEvent e) {
	}
}
