package jasymca;

import java.util.Stack;

import org.netlib.lapack.DGELSY;
import org.netlib.util.intW;

class LambdaLINLSTSQ extends LambdaAlgebraic{
 public int lambda(Stack st) throws ParseException, JasymcaException{
  try{
   Class c = Class.forName("org.netlib.lapack.DGELSY");
  }catch(ClassNotFoundException ce){
   throw new JasymcaException("JLAPACK not available.");
  }
  int narg = getNarg( st );
  if(narg!=2)
   throw new JasymcaException("Usage: LINLSTSQ( A, b ).");
  Matrix mx = new Matrix( getAlgebraic(st) );
  Algebraic b_in= getAlgebraic(st);
  Matrix bx = (b_in instanceof jasymca.Vektor ? Matrix.column((Vektor)b_in) : new Matrix( b_in ) );
  double a[][] = mx.getDouble();
  int m = mx.nrow();
  int n = mx.ncol();
  double b[][] = bx.getDouble(Math.max(n,m),0);
  int nrhs = bx.ncol();
  int jpvt[] = new int[n];
  for(int i=0; i<n; i++)
   jpvt[i] = 1;
  double[] work = new double[1];
  // condition number rcond:
  // error = rcond * eps
  double rcond = 1.0e-8;//15;
  intW rank = new intW(0);
  intW info = new intW(1);
  int lwork = -1;
  if(m != bx.nrow())
   throw new JasymcaException("Matrix dimensions wrong.");
  // query call
  DGELSY.DGELSY(m, n, nrhs, a, b,jpvt,rcond,rank,work,lwork,info);
  lwork = (int)work[0];
  if( lwork <= 0 )
   throw new JasymcaException("jlapack DGELSY query failed.");
  // set up work area
  work = new double[lwork];
  DGELSY.DGELSY(m, n, nrhs, a, b,jpvt,rcond,rank,work,lwork,info);
  if( info.val != 0 )
   throw new JasymcaException("jlapack DGELSY failed.");
//		System.out.println("Rank R11:"+rank.val);
  st.push( new Matrix(b,n,nrhs).reduce() );
  return 0;
 }
}

/*
@GROUP
equations 
@SYNTAX
linsolve(matrix1, matrix2).
@DOC
Solving systems of linear equations 
@NOTES
@EXAMPLES
A=[2 3 1; 4 4 5; 2 9 3];
b=[0;3;1];
a=linsolve(A,b)
printf('%f\n',a)

Answer:
0.90909
@SEE
*/

