package jasymca;

import jasymca.Algebraic;
import jasymca.JasymcaException;
import jasymca.LambdaAlgebraic;
import jasymca.Matrix;
import jasymca.ParseException;
import jasymca.Vektor;

import java.util.Stack;

import org.netlib.lapack.DGESV;
import org.netlib.util.intW;

class LambdaLINSOLVE2 extends LambdaAlgebraic{
 public int lambda(Stack st) throws ParseException, JasymcaException{
  try{
   Class c = Class.forName("org.netlib.lapack.DGESV");
  }catch(ClassNotFoundException ce){
   throw new JasymcaException("JLAPACK not available.");
  }
  int narg = getNarg( st );
  if(narg!=2)
   throw new JasymcaException("Usage: LINSOLVE2( A, b ).");
  Matrix mx = new Matrix( getAlgebraic(st) );
  Algebraic b_in= getAlgebraic(st);
  Matrix bx = (b_in instanceof jasymca.Vektor ? Matrix.column((Vektor)b_in) : new Matrix( b_in ) );
  double a[][] = mx.getDouble();
  double b[][] = bx.getDouble();
  int n = mx.nrow();
  int nrhs = bx.ncol();
  int ipiv[] = new int[n];
  intW info = new intW(1);
  if(n != mx.ncol() || n != bx.nrow())
   throw new JasymcaException("Matrix must be square.");

  DGESV.DGESV(n, nrhs, a, ipiv, b, info);
  if( info.val != 0 )
   throw new JasymcaException("jlapack DGESV failed.");
  st.push( new Matrix(b).reduce() );
  return 0;
 }
}


/*
@GROUP
equations 
@SYNTAX
linsolve2(matrix, vector).
@DOC
Solving linear system using LAPACK   
@NOTES
@EXAMPLES
a=solve(x^2-2*x*b+b^2,x)
printf('%f\n',a)
Answer:
b
@SEE
*/

