package jasymca;

import jasymca.Algebraic;
import jasymca.JasymcaException;
import jasymca.Lambda;
import jasymca.Matrix;
import jasymca.ParseException;
import jasymca.Vektor;

import java.util.Stack;



/**An external function for computing a mesh of a matrix  */
class LambdaDIAG extends Lambda{
 public int lambda(Stack st) throws ParseException, JasymcaException{
  // Variable definition, store in Prefix format and return success
  int narg = getNarg( st );
  if(narg<1)
   throw new JasymcaException("Usage: DIAG( matrix, k ).");
  Algebraic x = getAlgebraic( st ).reduce();
  int k = 0;
  if(narg>1)
   k = getInteger( st );
  if( x.scalarq() ){
   x = new Vektor(new Algebraic[] {x});
  }
  if( x instanceof jasymca.Vektor ){ // Create Matrix
   Vektor xv = (Vektor)x;
   if(k>=0){
    Matrix m = new Matrix(xv.length()+k, xv.length()+k);
    for(int i=0; i<xv.length(); i++){
     m.set(i, i+k, xv.get(i));
    }
    st.push( m );
   }else{
    Matrix m = new Matrix(xv.length()-k, xv.length()-k);
    for(int i=0; i<xv.length(); i++){
     m.set(i-k, i, xv.get(i));
    }
    st.push( m );
   }
  }else if (x instanceof jasymca.Matrix ){ // Extract vector
   Matrix xm = (Matrix)x;
   if(k>=0 && k<xm.ncol()){
    Algebraic a[] = new Algebraic[xm.ncol()-k];
    for(int i=0; i<a.length; i++){
     a[i] = xm.get(i, i+k);
    }
    st.push(new Vektor(a));
   }else if(k<0 && (-k)<xm.nrow()){
    Algebraic a[] = new Algebraic[xm.nrow()+k];
    for(int i=0; i<a.length; i++){
     a[i] = xm.get(i-k, i);
    }
    st.push(new Vektor(a));
   }else
    throw new JasymcaException("Argument k to DIAG out of range.");

  }else
   throw new JasymcaException("Argument to DIAG must be vector or matrix.");
  return 0;
 }
}



/*
@GROUP
Matrix
@SYNTAX
answer = diag (value)
@DOC
Returns the diagnoal of a matrix
@EXAMPLES
<programlisting>
diag([1,2,3;4,5,6;7,8,9]  
[1,
 5,
 9]
 
 diag([2,3,4;5,6,7],1]
 [3,
  7]
</programlisting>
@NOTES
.
@SEE
sum, trace, zeros, ones
*/
