package jasymca;

import jasymca.JasymcaException;
import jasymca.Lambda;
import jasymca.Matrix;
import jasymca.ParseException;

import java.util.Stack;


/**An external function for creating matrices that consist of zeros          *
 * everywhere except in the diagonal. The diagonal consists of ones.         *                        
 * (e.g.: eye(3) will return a 3-by-3 matrix [1,0,0;0,1,0;0,0,1],            *
 *  eye(4,3) will return a 4-by-3 matrix with diagonal set to 1              */


class LambdaEYE extends Lambda{
 public int lambda(Stack st) throws ParseException, JasymcaException{
  // Variable definition, store in Prefix format and return success
  int narg = getNarg( st );
  if(narg<1)
   throw new JasymcaException("Usage: EYE( nrow, ncol ).");
  int nrow = getInteger( st );
  int ncol = nrow;
  if(narg>1)
   ncol = getInteger( st );
  st.push( Matrix.eye(nrow, ncol).reduce() );
  return 0;
 }
}



/*
@GROUP
matrix
@SYNTAX
matrix = eye(size)
@DOC
returns a matrix with the specified size.
@NOTES
@EXAMPLES
<programlisting>
eye(2) = [1,0;0,1]
eye(3) = [1,0,0;0,1,0;0,0,1]
</programlisting>
@SEE
ones, zeros
*/
