package jasymca;

import jasymca.JasymcaException;
import jasymca.Lambda;
import jasymca.Matrix;
import jasymca.ParseException;
import jasymca.Zahl;

import java.util.Stack;

/**An external function for creating matrices that consist of ones           */
/* (e.g.: ones(2) will return a 2-by-2 matrix [1,2;1,1],                     *
 *  ones(4,3) will return a 4-by-3 matrix of ones                            */


class LambdaONES extends Lambda{
 public int lambda(Stack st) throws ParseException, JasymcaException{
  // Variable definition, store in Prefix format and return success
  int narg = getNarg( st );
  if(narg<1)
   throw new JasymcaException("Usage: ONES( nrow, ncol ).");
  int nrow = getInteger( st );
  int ncol = nrow;
  if(narg>1)
   ncol = getInteger( st );
  st.push( new Matrix(Zahl.ONE, nrow, ncol).reduce() );
  return 0;
 }
}


/*
@GROUP
matrix
@SYNTAX
ones(sizex, sizey)
ones(n,m,k,...)
@DOC
Returns a matrix of ones.
@NOTES
@EXAMPLES
ISNUMERIC(2,3) = [1,1,1;1,1,1]
ISNUMERIC(3,4) = [1,1,1,1;1,1,1,1;1,1,1,1]
@SEE
zeros

*/
