package jasymca;

import java.util.Stack;

import org.netlib.lapack.DGEQRF;
import org.netlib.lapack.DORGQR;
import org.netlib.lapack.DORMQR;
import org.netlib.util.intW;

class LambdaQR extends LambdaAlgebraic{
 public int lambda(Stack st) throws ParseException, JasymcaException{
  try{
   Class c = Class.forName("org.netlib.lapack.DGEQRF");
  }catch(ClassNotFoundException ce){
   throw new JasymcaException("JLAPACK not available.");
  }
  int narg = getNarg( st );
  if(narg!=1)
   throw new JasymcaException("Usage: QR( A ).");
  Matrix mx = new Matrix( getAlgebraic(st) );
  double a[][] = mx.getDouble();
  int m = mx.nrow(), n = mx.ncol();
  if(n>m)
   throw new JasymcaException("QR-factorization requires n<=m");
  double[] tau = new double[Math.min(m,n)];
        int lwork = n;
  double work[] = new double[lwork];
  intW info = new intW(1);
  DGEQRF.DGEQRF(m,n,a, tau,work,lwork,info);
  if( info.val != 0 )
   throw new JasymcaException("jlapack DGEQRF failed.");
  int k = Math.min(n,m);
  double c[][] = mx.getDouble();
  // Calculates R = Q^T * A --> c
  DORMQR.DORMQR("L","T",m,n,k,a,tau,c,work,lwork,info);
  if( info.val != 0 )
   throw new JasymcaException("jlapack DORMQR failed.");
  // Calculates Q --> a
  DORGQR.DORGQR(m,n,k,a,tau,work,lwork,info);
  if( info.val != 0 )
   throw new JasymcaException("jlapack DORGQR failed.");
  // force c into triangular form
  for(int i=0; i<c.length; i++)
   for(int j=0; j<i && j<c[0].length;j++)
    c[i][j] = 0.0;

  if( length==2 ){
   st.push( (new Matrix(a)).reduce() );
   st.push( (new Matrix(c)).reduce() );
   length = 1;
     }else
      throw new JasymcaException("Wrong number of output variables: "+length);
     return 0;
 }
}


/*
@GROUP
matrix
@SYNTAX
qr(matrix)
@DOC
Calculates the QR decomposition of matrix.
@NOTES
   For an m-by-n matrix A with m >= n, the QR decomposition is an m-by-n
   orthogonal matrix Q and an n-by-n upper triangular matrix R so that
   A = Q*R.

   The QR decompostion always exists, even if the matrix does not have
   full rank, so the constructor will never fail.  The primary use of the
   QR decomposition is in the least squares solution of nonsquare systems
   of simultaneous linear equations.  This will fail if isFullRank()
   returns false.
@EXAMPLES
@SEE
lu
*/
