package jasymca;

import jasymca.JasymcaException;
import jasymca.Lambda;
import jasymca.Matrix;
import jasymca.ParseException;

import java.util.Stack;


/**An external function for creating matrices that consist of zeros          *
* (e.g.: zeros(2) will return a 2-by-2 matrix [0,0;0,0],                     *
*  zeros(4,3) will return a 4-by-3 matrix of zeros                           */


class LambdaZEROS extends Lambda{
 public int lambda(Stack st) throws ParseException, JasymcaException{
  // Variable definition, store in Prefix format and return success
  int narg = getNarg( st );
  if(narg<1)
   throw new JasymcaException("Usage: ZEROS( nrow, ncol ).");
  int nrow = getInteger( st );
  int ncol = nrow;
  if(narg>1)
   ncol = getInteger( st );
  st.push( new Matrix(nrow, ncol).reduce() );
  return 0;
 }
}


/*
@GROUP
matrix
@SYNTAX
zeros(sizey, [sizex])
zeros(n,m,k,l...)
@DOC
Returns a matrix filled with zeros.
@NOTES
@EXAMPLES
<programlisting>
zeros(2,2) = [0,0;0,0]
zeros(2,3) = [0,0,0;0,0,0]
</programlisting>
@SEE
ones
*/
