package jasymca;

import java.util.Stack;
import jasymca.*;
import jhplot.math.num.random.RandomRNG;

class LambdaRANDOM extends Lambda {
	public int lambda(Stack st) throws ParseException, JasymcaException {
		int narg = getNarg(st);
		if (narg != 1 && narg != 2)
			throw new ParseException(
					"requires 1 or 2 arguments: Dimension X and Y");

		if (narg == 2) {
			Algebraic x = getAlgebraic(st);
			Algebraic y = getAlgebraic(st);

			// generate matrix
			if (x instanceof jasymca.Zahl && y instanceof jasymca.Zahl) {

				int j1 = (int) ((Zahl) x).unexakt().real;
				int j2 = (int) ((Zahl) y).unexakt().real;

				RandomRNG p = new RandomRNG();

				double[][] d = new double[j1][j2];
				for (int i = 0; i < j1; i++) {
					for (int j = 0; j < j2; j++) {
						d[i][j] = p.nextRandomNumber();
					}
				}
				Matrix m = new Matrix(d);
				st.push(m);
			}
		}

		if (narg == 1) {
			Algebraic x = getAlgebraic(st);

			// generate matrix
			if (x instanceof jasymca.Zahl) {

				int j1 = (int) ((Zahl) x).unexakt().real;
				RandomRNG p = new RandomRNG();

				double[] d = new double[j1];
				for (int i = 0; i < j1; i++) {
					d[i] = p.nextRandomNumber();

				}
				Vektor m = new Vektor(d);
				st.push(m);
			}
		}

		return 0;
	}

}

/*
 * @GROUP random
 * 
 * @SYNTAX random(x, y)
 * 
 * @DOC Creates a matrix or vector with random numbers in the range [0-1; If x,y
 * are given , creates a matrix with the dimension x (rows) and y (columns). If
 * only x is given, creates a vector of size "x"
 * 
 * @EXAMPLES <programlisting> random(100) % creates a vector with size 100.
 * random(50,60) % matrix 50x60 will be filled. </programlisting>
 * 
 * @NOTES
 * 
 * @SEE beta
 */
