package jasymca;

import jasymca.*;
import java.util.Stack;
import java.io.*;

import jhplot.stat.Statistics;


class LambdaCOV extends Lambda{
 public int lambda(Stack st) throws ParseException, JasymcaException{
  int narg = getNarg( st );
  if(narg!=2)
   throw new ParseException("Wrong number of arguments. Should be either 2 matricies or 2 vectors");
  Algebraic a = getAlgebraic( st );
  Algebraic b = getAlgebraic( st );

  if( a instanceof jasymca.Vektor && b instanceof jasymca.Vektor ){
     Vektor m1 = new Vektor( a );
     double v1[] = m1.getDouble();
     Vektor m2 = new Vektor( b );
     double v2[] = m2.getDouble();
     double m= Statistics.covariance(v1,v2);
     st.push( new Unexakt((double)m)  );
  }


   if( a instanceof jasymca.Matrix && b instanceof jasymca.Matrix ){
     Matrix m1 = new Matrix( a );
     double v1[][] = m1.getDouble();
     Matrix m2 = new Matrix( b );
     double v2[][] = m2.getDouble();
     double m[][]= Statistics.covariance(v1,v2);
     st.push( new Matrix(m) );
  }

  return 0;
 }
}


/*
@GROUP
statistics
@SYNTAX
cov([2,3,8,...],[1,3,4,...])
@DOC
Calculates the covariance for 2 matricies or 2 vectors.
@EXAMPLES
<programlisting>
cov([3,4,5],[3,4,5])
</programlisting>
@SEE
mean, std
*/

