package jasymca;

import jasymca.Algebraic;
import jasymca.JasymcaException;
import jasymca.LambdaAlgebraic;
import jasymca.Unexakt;
import jasymca.Zahl;

class LambdaATAN extends LambdaAlgebraic{
 public LambdaATAN(){
  diffrule = "1/(1+x^2)";
  intrule = "x*atan(x)-1/2*log(1+x^2)";
  trigrule = "-i/2*log((1+i*x)/(1-i*x))";
 }
 Zahl f( Zahl x) throws JasymcaException{
  Unexakt z = x.unexakt();
  if(z.imag == 0.)
   return new Unexakt(Math.atan(z.real));
  return (Zahl)evalx(trigrule, z);
 }
 Algebraic f_exakt(Algebraic x) throws JasymcaException{
  if(x.equals(Zahl.ZERO))
   return Zahl.ZERO;
  return null;
 }
}

/*
@GROUP
trigonometric
@SYNTAX
atan(angle)
@DOC
Calculates the inverse tangent of angle.
@EXAMPLES
<programlisting>
atan(1) = 0.7853981633974483
atan(0) = 0
</programlisting>
@SEE
atan2, tan
*/
