
/**
 * Class for prompt
 */
package jmathlab;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.*;
import java.util.Vector;
import javax.swing.*;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

/**
 * Class implementing a console style window It needs the ConsoleKeyHandler
 * class to work
 */
public class Console extends JTextPane {

	private JasymcaSyntaxDocument doc;

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	int commandNo;

	/** store for previous commands */
	private Vector<String> previousCommands;

	/** The position of the start of the line */
	int lineStart;

	private JPrompt callerClass;

	
	final private String welcome="% jMathLab. Type 'help' to learn more\n";
	
	/** Event Handler used for handling key events */
	// public KeyListener keyHandler;

	/** Construct the console */
	public Console(JPrompt gui) {

		commandNo = 0;
		previousCommands = new Vector<String>(10, 10);
		lineStart = getText().length() + 1;

		callerClass = gui;

	 	KeyListener keyHandler = new ConsoleKeyHandler();
	 	addKeyListener(keyHandler);
		setLayout(new BorderLayout());
		doc = new JasymcaSyntaxDocument();
		setDocument(doc);
		setCaretColor(Color.red);
		setFont(jmatmain.SetEnv.globalFont);
                setText(welcome);
	}

	public void addKeyListener(KeyListener keyHandler) {
		// System.out.println("KEYLISTENER");
		//System.out.println("KEY " + keyHandler);
		super.addKeyListener(keyHandler);
	}

	/** display the previous command in the list */
	public void prevCommand() {
		commandNo--;

		String text = "";
		if (commandNo >= 0 && commandNo < previousCommands.size()) {
			text = ((String) previousCommands.elementAt(commandNo)).trim();
		} else if (commandNo < 0) {
			text = "";
			commandNo = -1;
		}

		// replace current command with previous command
		String textArea = getText();
		int pos1 = textArea.lastIndexOf("> ") + 2;
		setText(textArea.substring(0, pos1) + text);

		// set cursor at the end of the text area
		setCaretPosition(getText().length());
	}

	/** display the next command in the list */
	public void nextCommand() {
		commandNo++;

		String text = "";
		if (commandNo >= 0 && commandNo < previousCommands.size()) {
			text = ((String) previousCommands.elementAt(commandNo)).trim();
		} else if (commandNo >= previousCommands.size()) {
			text = "";
			commandNo = previousCommands.size();
		}

		// replace current command with next command
		String textArea = getText();
		int pos1 = textArea.lastIndexOf("> ") + 2;
		setText(textArea.substring(0, pos1) + text);

		// set cursor at the end of the text area
		setCaretPosition(getText().length());
	}

	/**
	 * Check that the cursor hasn't moved beyond the start of the line
	 */
	public void checkPosition() {
		if (getCaretPosition() < lineStart)
			setCaretPosition(lineStart);
	}

	/** Go to the home position */
	public void home() {
		setCaretPosition(lineStart - 1);
	}

	/** Go to the end of the line */
	public void end() {
		setCaretPosition(getText().length());
	}

	/** Interpret the current line */
	public void interpretLine() {
		// get the text on the current line
		String text = getText();
		String inputString = text.substring(text.lastIndexOf("> ") + 2,
				text.length());

		/* exit application */
		if (inputString.equals("quit") || inputString.equals("exit") || inputString.equals("clear")) {
			callerClass.clear();
		}

		append("\n");
		// System.out.println("Interpret="+inputString);

		callerClass.interpretLine(inputString);
		previousCommands.addElement(inputString);
		commandNo = previousCommands.size();
	}

	/** Display the command prompt */
	public void displayPrompt() {
		append("> ");
		lineStart = getText().length();
		setCaretPosition(lineStart);
	}

	// Display a new line*/
	public void newLine() {
		append("\n");
	}

	/** Display some text on a new line */
	public void displayText(String text) {
                doc.setHighlighting(false);
	        	append(text + "\n");
                doc.setHighlighting(true);
 
	}

	public JTextPane getComponent() {
		return this;
	}

	
	/**
	 * Restart
	 */
	public void clear() {
		setText(welcome);
		displayPrompt();
	}
	
	/**
	 * Append text
	 * 
	 * @param s
	 */
	public void append(String s) {
		try {
			Document doc = getDocument();
			doc.insertString(doc.getLength(), s, null);
		} catch (BadLocationException exc) {
			exc.printStackTrace();
		}
	}

}
