package jmatmain;

import java.awt.BorderLayout;
import javax.swing.*;
import java.awt.*;
import java.io.*;
import java.awt.event.*;
import jasymca.JasymcaConsoleText;
import java.io.InputStream;
import java.io.ByteArrayInputStream;

/**
 * Application. All m files included.
 * 
 * @author sergei
 * 
 */
public class MainAll {

	private static JTabbedPane jtab;
	private static mainGUI maingui;
	private static OutText jconsole;
	private static OutSysIn jconsolesys;
	private static Thread trun;
	private static ConsoleService jasymcaConsole;
	
	public static void main(String[] args) {

		// command line
		if (args.length != 0) {
			String filename = args[0];
			JasymcaConsoleText jasymcaConsoletext = new JasymcaConsoleText(System.in,System.out);
			String s = getContents(new File(filename));
			try {
				InputStream is = new ByteArrayInputStream(s.getBytes("UTF-8"));
				jasymcaConsoletext.readStream(is);
			} catch (Exception ie) {
				System.out.println(ie.toString());
			}
			// System.out.println("Done!");
			// System.exit(1);
			return;
		}

		JButton button = new JButton("Run");
		JButton reload = new JButton("Reload");

		button.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String s = maingui.getEditor().getText();
				evalSymb(s);
			}
		});

		reload.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
                                jasymcaConsole.clear(); 
				jconsole.setText("");
				jtab.setSelectedIndex(0);
			}
		});

		JFrame frame = new JFrame("JMathLab");
		frame.setPreferredSize(new Dimension(500, 500));

		getGUI();
		frame.add(jtab, BorderLayout.CENTER);
		frame.add(button, BorderLayout.NORTH);
		frame.add(reload, BorderLayout.SOUTH);

		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.pack();
		frame.setVisible(true);

	}

	private static void ErrorMessage(String a) {

		JOptionPane dialogError = new JOptionPane();
		JOptionPane.showMessageDialog(dialogError, a, "Error",
				JOptionPane.ERROR_MESSAGE);
	}

	

	public static void evalSymb(final String s) {
		trun = new Thread() {
			public void run() {

				try {
					String sout=jasymcaConsole.exec(s);
					jconsole.setText(sout);
					jtab.setSelectedIndex(1);
                                        String serr=jasymcaConsole.getError(); 
                                        if (serr.length()>1) ErrorMessage(serr);
				} catch (Exception ie) {
					ErrorMessage(ie.toString());
				}

			};
		};
		trun.start();

	}

	public static void getGUI() {

		jconsole = new OutText();
		jconsole.setFont(SetEnv.globalFont);
		JScrollPane scrollingArea = new JScrollPane(jconsole);
		maingui = new mainGUI(true);
		jtab = new JTabbedPane();
		jtab.addTab("Editor", maingui);
		jtab.addTab("Output", scrollingArea);
		jasymcaConsole = new ConsoleService();

	}

	/**
	 * Fetch the entire contents of a text file, and return it in a String. This
	 * style of implementation does not throw Exceptions to the caller.
	 * 
	 * @param aFile
	 *            is a file which already exists and can be read.
	 */
	static public String getContents(File aFile) {
		// ...checks on aFile are elided
		StringBuilder contents = new StringBuilder();

		try {
			// use buffering, reading one line at a time
			// FileReader always assumes default encoding is OK!
			BufferedReader input = new BufferedReader(new FileReader(aFile));
			try {
				String line = null; // not declared within while loop
				/*
				 * readLine is a bit quirky : it returns the content of a line
				 * MINUS the newline. it returns null only for the END of the
				 * stream. it returns an empty String if two newlines appear in
				 * a row.
				 */
				while ((line = input.readLine()) != null) {
					contents.append(line);
					contents.append(System.getProperty("line.separator"));
				}
			} finally {
				input.close();
			}
		} catch (IOException ex) {
			ex.printStackTrace();
		}

		return contents.toString();
	}
}
