package jmatmain;

import java.awt.Font;
import java.awt.Color;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Dimension;
import javax.swing.BorderFactory;
import javax.swing.JScrollPane;
import javax.swing.JApplet;
import javax.swing.JButton;
import javax.swing.JOptionPane;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;



/**
 * Main class to load
 * 
 * @author sergei chekanov
 * 
 */
public class MainAppletInline extends JApplet {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private JSplitPane jsplit;
	private mainGUI maingui;
	private int asize = 500;
	private String input = ""; // source code input
	private String control = "none"; // control. mamber is complete access
	private int fontsize = 12;
	private int nrlines = 1;
        private OutText jconsole;
	private String eol = System.getProperty("line.separator");
	private JButton jrun;
	private boolean editMode;
        private final Color cborder= new Color(66,194,70); // border around frame in edit mode
        private Thread trun;
        private ConsoleService jasymcaConsole;

 
	public void init() {

		editMode = false;
		final String paramSize = getParameter("asize");
		if (paramSize != null) {
			try {
				asize = Integer.parseInt(paramSize);
			} catch (final NumberFormatException e) {
			}
		}

                input = getParameter("input");
                input = input.replace("&newline;", "\n");
                input = input.replace("&quote;", "\"");

		// input = getParameter("input");
		// input = input.replace("[%]", "\n");
		// String[] number_array = input.split("stop|\\n");
		String[] number_array = input.split(eol);
		nrlines = number_array.length;

		control = getParameter("control");
		if (control.indexOf("edit") > -1)
			editMode = true;

		SetEnv.outputString = "";
		final String fontSize = getParameter("fontsize");
		if (paramSize != null) {
			try {
				fontsize = Integer.parseInt(fontSize);
				String name = SetEnv.globalFont.getFamily();
				SetEnv.globalFont = new Font(name, Font.BOLD, fontsize);
			} catch (final NumberFormatException e) {
			}

		}

		
	
		//Execute a job on the event-dispatching thread; creating this applet's GUI.
        try {
            SwingUtilities.invokeAndWait(new Runnable() {
                public void run() {
                	getGUI(); 
                }
            });
        } catch (Exception e) {
            System.err.println("createGUI didn't complete successfully");
        }
    
			
		

	}

	

	/**
	 * Make GUI
	 * 
	 * @return
	 */
	public void getGUI() {

               jconsole = new OutText();
               jconsole.setFont(SetEnv.globalFont);

                JScrollPane scrollingArea = new JScrollPane(jconsole);
                scrollingArea.getVerticalScrollBar().setPreferredSize(new Dimension(10, 0));
                scrollingArea.getHorizontalScrollBar().setPreferredSize(new Dimension(0, 10));

		jconsole.setEditable(false);
		jconsole.setBorder(javax.swing.BorderFactory.createEmptyBorder());

		// jconsole.setLayout(new BorderLayout());
		maingui = new mainGUI(false);
		maingui.getEditor().setHighlightCurrentLine(false);
		maingui.getEditor().setEditable(false);
		// maingui.setLayout(new BorderLayout());
		maingui.setBorder(javax.swing.BorderFactory.createEmptyBorder());

		maingui.getScrollPane().setHorizontalScrollBarPolicy(
				JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		maingui.getScrollPane().setVerticalScrollBarPolicy(
				JScrollPane.VERTICAL_SCROLLBAR_NEVER);
		maingui.getScrollPane().setBorder(
				javax.swing.BorderFactory.createEmptyBorder());

		// jconsole.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		// jconsole.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);

		jsplit = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, maingui, scrollingArea);
		jsplit.setBorder(javax.swing.BorderFactory.createEmptyBorder());

		// allow to edit
		if (editMode == true) {
			maingui.getEditor().setEditable(true);
			jsplit.setBorder(BorderFactory.createLineBorder(cborder)); 
			maingui.setBorder(BorderFactory.createLineBorder(cborder));
		}

		// jsplit.setDividerLocation(150);
		jsplit.setBackground(Color.white);
		jsplit.setResizeWeight(0.6);
		jsplit.setDividerLocation(0.6);
		jsplit.setDividerSize(4);
		
		
		
		
		
		maingui.getEditor().setText(input);
		jconsole.setText("");
		// jtab.setSelectedIndex(0);
		add(jsplit, BorderLayout.CENTER);


                        jrun = new JButton("Run");
                        jrun.setOpaque(true);
                        jrun.setBackground(new Color(197, 237, 250));
                        if (control.indexOf("member") == -1) jrun.setBackground(new Color(245,172,172));

                        jrun.addActionListener(new ActionListener() {
                                public void actionPerformed(ActionEvent e) {
                                       if (control.indexOf("member") > -1) {
                                        String s = maingui.getEditor().getText();
                                            evalSymb(s);
                                        } else ErrorMessage("Non-members cannot run this example!");
                                }
                        });

                        add(jrun, BorderLayout.EAST);



		

		return;
	}

	// -----------------------------------------------------------------------
	public void start() {
		// maingui.getEditor().setText(input);
		// jtab.setSelectedIndex(0);
		SetEnv.outputString = "";
		jconsole.setText("");
		// jtab.setSelectedIndex(0);

	}

	// -----------------------------------------------------------------------

	
	/**
	 * Stop calculations
	 */

/*
	public void stop() {
	   	    trun.interrupt();
	        trun = null;
	        /// System.out.println("stopping... ");
		
		
	}

	
	public void destroy() {
		
		if(trun != null){
            trun = null;
        }
	}
*/

	private void ErrorMessage(String a) {

		JOptionPane dialogError = new JOptionPane();
		JOptionPane.showMessageDialog(dialogError, a, "Error",
				JOptionPane.ERROR_MESSAGE);
	}

	
	
	
            public void evalSymb(final String s) {
             trun = new Thread("Evaluate symbolic calculations") {
                    public void run() {

                       try
                                {
                                jasymcaConsole = new ConsoleService();
                                jasymcaConsole.clear(); 
                                jconsole.setText("");
                                String sout=jasymcaConsole.exec(s);
                                String serr=jasymcaConsole.getError();
                                if (serr.length()>1) ErrorMessage(serr);

                                 String[] number_array = sout.split(eol);
                                 int outlines = number_array.length;
                                 if (outlines <= nrlines) {
                                           jconsole.setText(sout);
                                        } else {
                                                OutDialog da = new OutDialog(jconsole);
                                                da.setText(sout);
                                                da.setVisible(true);

                                        }
 

                                 }catch (Exception ie)
                                {
                                  ErrorMessage(ie.toString());
                               }






                       };
                      };
       trun.start();

       }


	/*
	 * 
	 * 
	 * //
	 * -----------------------------------------------------------------------
	 * 
	 * public void destroy() {
	 * 
	 * if (jconsole != null) jconsole.clear(); jconsole = null; maingui = null;
	 * jtab = null;
	 * 
	 * }
	 * 
	 * 
	 * private void ErrorMessage(String a) {
	 * 
	 * JOptionPane dialogError = new JOptionPane();
	 * JOptionPane.showMessageDialog(dialogError, a, "Error",
	 * JOptionPane.ERROR_MESSAGE); }
	 * 
	 * 
	 * /* // soft refresh JythoShell in thread (new instance) public void
	 * iniPython() { Thread t = new Thread("iniPython") { public void run() {
	 * jconsole = new JyShell(); jtab.repaint(); }; }; t.start(); }
	 */

}
