// * This code is licensed under:
// * jeHEP License, Version 1.0
// * - for license details see http://jehep.sourceforge.net/license.html 
// *
// * Copyright (c) 2005 by S.Chekanov (chekanov@mail.desy.de). 
// * All rights reserved.
package jmatmain;

import javax.swing.JPanel;
import java.awt.BorderLayout;
import java.awt.Font;


import org.fife.ui.rsyntaxtextarea.modes.MatlabTokenMaker; 
import org.fife.ui.rtextarea.*;
import org.fife.ui.rsyntaxtextarea.*;
import org.fife.ui.rsyntaxtextarea.RSyntaxDocument;
import org.fife.ui.rsyntaxtextarea.RSyntaxTextArea;
import org.fife.ui.rsyntaxtextarea.Style;
import org.fife.ui.rsyntaxtextarea.SyntaxScheme;

//import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.text.*;


//import org.fife.ui.rtextarea.*;
//import org.fife.ui.rsyntaxtextarea.*;


/**
 * This is editor
 * 
 * @author sergei
 * 
 */

public class mainGUI extends JPanel {

        private RSyntaxTextArea textArea;
         private RTextScrollPane sp;

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	public mainGUI(boolean isPopup) {

      setLayout(new BorderLayout());

      RSyntaxDocument syns= new RSyntaxDocument(SyntaxConstants.SYNTAX_STYLE_NONE);
      syns.setSyntaxStyle( new MatlabTokenMaker() );
      textArea = new RSyntaxTextArea(syns);
      textArea.setCodeFoldingEnabled(false);
      textArea.setAntiAliasingEnabled(true);
      textArea.setFont(SetEnv.globalFont);   
      sp = new RTextScrollPane(textArea);
      sp.setFoldIndicatorEnabled(false);


// Add an item to the popup menu that opens the file whose name is
       // specified at the current caret position.
      if (isPopup==true) {
      JPopupMenu popup = textArea.getPopupMenu();
      popup.addSeparator();
      popup.add(new JMenuItem(new OpenFileAction()));
      textArea.setCodeFoldingEnabled(true);
       sp.setFoldIndicatorEnabled(true);
      }


      add(sp);

    }

/**
 *     * Loads a file's contents into the text area, or displays an error message
 *         * if the file does not exist.
 *             * 
 *                 * @param file
 *                     *           The file to load.
 *                         */
   public void loadFile(File file) {

//      System.out.println("DEBUG: " + file.getAbsolutePath());
      if (file.isDirectory()) { // Clicking on a space character
         JOptionPane.showMessageDialog(this, file.getAbsolutePath()
               + " is a directory", "Error", JOptionPane.ERROR_MESSAGE);
         return;
      } else if (!file.isFile()) {
         JOptionPane.showMessageDialog(this, "No such file: "
               + file.getAbsolutePath(), "Error", JOptionPane.ERROR_MESSAGE);
         return;
      }

      try {
         BufferedReader r = new BufferedReader(new FileReader(file));
         textArea.read(r, null);
         r.close();
      } catch (IOException ioe) {
         ioe.printStackTrace();
         UIManager.getLookAndFeel().provideErrorFeedback(textArea);
      }

   }


/**
 *     * Changes the styles used by the editor via an XML file specification. This
 *         * method is preferred because of its ease and modularity.
 *             */
   private void changeStyleViaThemeXml() {
      try {
         Theme theme = Theme.load(getClass().getResourceAsStream(
               "/eclipse_theme.xml"));
         theme.apply(textArea);
      } catch (IOException ioe) { // Never happens
         ioe.printStackTrace();
      }
   }

   /**
 *     * Set the font for all token types.
 *         * 
 *             * @param textArea The text area to modify.
 *                 * @param font The font to use.
 *                     */
   public static void setFont(RSyntaxTextArea textArea, Font font) {
      if (font != null) {
         SyntaxScheme ss = textArea.getSyntaxScheme();
         ss = (SyntaxScheme) ss.clone();
         for (int i = 0; i < ss.getStyleCount(); i++) {
            if (ss.getStyle(i) != null) {
               ss.getStyle(i).font = font;
            }
         }
         textArea.setSyntaxScheme(ss);
         textArea.setFont(font);
      }
   }


  /**
    * An action that gets the filename at the current caret position and tries
    * to open that file. If there is a selection, it uses the selected text as
    * the filename.
    */
   private class OpenFileAction extends TextAction {

      public OpenFileAction() {
         super("Open File");
      }

      public void actionPerformed(ActionEvent e) {

         JTextComponent tc = getTextComponent(e);
         String filename = null;

         // Get the name of the file to load. If there is a selection, use
         // that as the file name, otherwise, scan for a filename around
         // the caret.
         try {
            int selStart = tc.getSelectionStart();
            int selEnd = tc.getSelectionEnd();
            if (selStart != selEnd) {
               filename = tc.getText(selStart, selEnd - selStart);
            } else {
               filename = getFilenameAtCaret(tc);
            }
         } catch (BadLocationException ble) {
            ble.printStackTrace();
            UIManager.getLookAndFeel().provideErrorFeedback(tc);
            return;
         }

         loadFile(new File(filename));

      }

      /**
       * Gets the filename that the caret is sitting on. Note that this is a
       * somewhat naive implementation and assumes filenames do not contain
       * whitespace or other "funny" characters, but it will catch most common
       * filenames.
       * 
       * @param tc
       *           The text component to look at.
       * @return The filename at the caret position.
       * @throws BadLocationException
       *            Shouldn't actually happen.
       */
      public String getFilenameAtCaret(JTextComponent tc) throws BadLocationException {
         int caret = tc.getCaretPosition();
         int start = caret;
         Document doc = tc.getDocument();
         while (start > 0) {
            char ch = doc.getText(start - 1, 1).charAt(0);
            if (isFilenameChar(ch)) {
               start--;
            } else {
               break;
            }
         }
         int end = caret;
         while (end < doc.getLength()) {
            char ch = doc.getText(end, 1).charAt(0);
            if (isFilenameChar(ch)) {
               end++;
            } else {
               break;
            }
         }
         return doc.getText(start, end - start);
      }

      public boolean isFilenameChar(char ch) {
         return Character.isLetterOrDigit(ch) || ch == ':' || ch == '.'
               || ch == File.separatorChar;
      }

   }









 /**
 * Return editor
 *                 
 *                 @return
 **/
        public RTextScrollPane  getScrollPane() {

                return sp;
        }


	/**
	 * Return editor
	 * 
	 * @return
	 */
	public RSyntaxTextArea  getEditor() {

		return textArea;
	}


}
